/*
 * Copyright (c) 2006, Matt Whitlock and WhitSoft Development
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the names of Matt Whitlock and WhitSoft Development nor the
 *     names of their contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "userdb.h"

bool UserDB::Add(const char *pszUsername, bool bGroup)
{
	if (_users.find(pszUsername) == _users.end()) {
		_users.insert(std::make_pair(pszUsername, USERDBRECORD()));
		map_type::iterator it = _users.find(pszUsername);
		if (it != _users.end()) {
			it->second.bGroup = bGroup;
			it->second.bLoginForbidden = FALSE;
			it->second.szPassiveIP[0] = 0x00;
			it->second.szYubikeyID[0] = 0x00;
#if (defined SLIMFTPD_VARIABLES) && (defined SLIMFTPD_VARIABLES_USERNAME)
//			VFS *vfs = GetVFS(pszUsername);
			VFS *vfs = &it->second.vfs;
			if (vfs != NULL) 
				vfs->strUsername.assign(pszUsername);
//			PermDB *perms = GetPermDB(pszUsername);
			PermDB *perms = &it->second.perms;
			if (perms != NULL) 
				perms->strUsername.assign(pszUsername);
#endif
		}
		return true;
	} else {
		return false;
	}
}

bool UserDB::SetPassword(const char *pszUsername, const char *pszPassword)
{
	map_type::iterator it = _users.find(pszUsername);
	if (it != _users.end()) {
		it->second.strPassword = pszPassword;
		return true;
	}
	return false;
}

bool UserDB::AddParent(const char *pszUsername, const char *pszParent)
{
#ifdef DEBUG
printf("AddParent('%s', '%s')\n", pszUsername, pszParent);
#endif
	map_type::iterator parent = _users.find(pszParent);
	if (parent == _users.end())
		return false;
	SetLoginForbidden(pszUsername, GetLoginForbidden(pszParent));
	char * psz = GetPassiveIP(pszParent);
	if (psz != NULL)
		SetPassiveIP(pszUsername, psz);
	GetVFS(pszParent)->CopyMountsTo(GetVFS(pszUsername));
	GetPermDB(pszParent)->CopyPermsTo(GetPermDB(pszUsername));
	return true;
}

bool UserDB::SetLoginForbidden(const char *pszUsername, bool bForbidden)
{
	map_type::iterator it = _users.find(pszUsername);
	if (it != _users.end()) {
		it->second.bLoginForbidden = bForbidden;
		return true;
	}
	return false;
}

VFS * UserDB::GetVFS(const char *pszUsername)
{
	map_type::iterator it = _users.find(pszUsername);
	if (it != _users.end()) {
		return &it->second.vfs;
	}
	return NULL;
}

PermDB * UserDB::GetPermDB(const char *pszUsername)
{
	map_type::iterator it = _users.find(pszUsername);
	if (it != _users.end()) {
		return &it->second.perms;
	}
	return NULL;
}

bool UserDB::CheckPassword(const char *pszUsername, const char *pszPassword)
{
	if (!CheckLoginAllowed(pszUsername))
		return FALSE;
	map_type::iterator it = _users.find(pszUsername);
	if (it == _users.end())
		return FALSE;
	if (it->second.szYubikeyID[0] == '\0')
		return (it->second.strPassword == pszPassword);
	else {
		const char *p = it->second.strPassword.c_str();
		const char *q = pszPassword;
		while (*p)
			if ((!(*q)) || (*p != *q))
				return FALSE;
			else {
				p++;
				q++;
			}
		// *q points to Yubikey OTP
		HMODULE yubilib = LoadLibrary("PkYubikeyDll.dll");
		if (!yubilib)
			return FALSE;
		typedef long (WINAPI *PPKYubikeyValidateOTP) (const char *Name, const char *OTP);
		bool Result = FALSE;
		PPKYubikeyValidateOTP PKYubikeyValidateOTP = (PPKYubikeyValidateOTP) GetProcAddress(yubilib, "PKYubikeyValidateOTP");
		if (PKYubikeyValidateOTP) {
			Result = !PKYubikeyValidateOTP(it->second.szYubikeyID, q);
		}
		FreeLibrary(yubilib);
		return Result;
	}
}

bool UserDB::GetLoginForbidden(const char *pszUsername)
{
	map_type::iterator it = _users.find(pszUsername);
	return (it == _users.end()) || (it->second.bLoginForbidden);
}

bool UserDB::GetGroup(const char *pszUsername)
{
	map_type::iterator it = _users.find(pszUsername);
	return (it != _users.end()) && (it->second.bGroup);
}

bool UserDB::CheckLoginAllowed(const char *pszUsername)
{
	return (!GetLoginForbidden(pszUsername)) && (!GetGroup(pszUsername));
}

bool UserDB::SetPassiveIP(const char *pszUsername, const char *pszPassiveIP)
{
	map_type::iterator it = _users.find(pszUsername);
	if (it != _users.end()) {
		strncpy(it->second.szPassiveIP, pszPassiveIP, sizeof(it->second.szPassiveIP)-1);
		return true;
	}
	return false;
}

char * UserDB::GetPassiveIP(const char *pszUsername)
{
	map_type::iterator it = _users.find(pszUsername);
	if (it != _users.end()) {
		if (strlen(it->second.szPassiveIP) > 0)
			return (it->second.szPassiveIP);
		else
			return NULL;
	}
	return NULL;
}

bool UserDB::SetYubikeyID(const char *pszUsername, const char *pszYubikey)
{
	map_type::iterator it = _users.find(pszUsername);
	if (it != _users.end()) {
		strncpy(it->second.szYubikeyID, pszYubikey, YUBIKEY_ID_LENGTH);
		it->second.szYubikeyID[YUBIKEY_ID_LENGTH-1] = '\0';
		return true;
	}
	return false;
}

